# Copyright (c) 2023 ING Analytics Wholesale Banking
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import warnings

import numpy as np
import pandas as pd

from popmon.base import Module
from popmon.utils import parallel


class ApplyFunc(Module):
    """This module applies functions to specified feature and metrics.

    Extra parameters (kwargs) can be passed to the apply function.
    """

    _input_keys = ("apply_to_key", "assign_to_key", "apply_funcs_key")
    _output_keys = ("store_key",)

    def __init__(
        self,
        apply_to_key,
        store_key: str = "",
        assign_to_key: str = "",
        apply_funcs_key: str = "",
        features=None,
        apply_funcs=None,
        metrics=None,
        msg: str = "",
    ) -> None:
        """Initialize an instance of ApplyFunc.

        :param str apply_to_key: key of the input data to apply funcs to.
        :param str assign_to_key: key of the input data to assign function applied-output to. (optional)
        :param str store_key: key of the output data to store in the datastore (optional)
        :param str apply_funcs_key: key of to-be-applied functions in data to store (optional)
        :param list features: list of features to pick up from input data and apply funcs to (optional)
        :param list metrics: list of metrics to apply funcs to (optional)
        :param str msg: message to print out at start of transform function. (optional)
        :param list apply_funcs: functions to apply (list of dicts):

          - 'func': function to apply
          - 'suffix' (string, optional): suffix added to each metric. default is function name.
          - 'prefix' (string, optional): prefix added to each metric.
          - 'features' (list, optional): features the function is applied to. Overwrites features above
          - 'metrics' (list, optional): metrics the function is applied to. Overwrites metrics above
          - 'entire' (boolean, optional): apply function to the entire feature's dataframe of metrics?
          - 'args' (tuple, optional): args for 'func'
          - 'kwargs' (dict, optional): kwargs for 'func'

        """
        super().__init__()
        self.apply_to_key = apply_to_key
        self.assign_to_key = assign_to_key or apply_to_key
        self.apply_funcs_key = apply_funcs_key
        self.store_key = store_key or self.assign_to_key

        self.features = features or []
        self.metrics = metrics or []
        self.msg = msg
        self.apply_funcs = []
        # import applied functions
        apply_funcs = apply_funcs or []
        for af in apply_funcs:
            self.add_apply_func(**af)

    def get_description(self):
        if len(self.apply_funcs) > 0:
            return " and ".join([x["func"].__name__ for x in self.apply_funcs])
        elif self.apply_funcs_key:
            return f"functions from arg '{self.apply_funcs_key}'"
        else:
            raise NotImplementedError

    def add_apply_func(
        self,
        func,
        suffix=None,
        prefix=None,
        metrics=None,
        features=None,
        entire=None,
        *args,
        **kwargs,
    ):
        """Add function to be applied to dataframe.

        Can call this function after module instantiation to add new functions.

        :param func: function to apply
        :param suffix: (string, optional) suffix added to each metric. default is function name.
        :param prefix: (string, optional) prefix added to each metric.
        :param features: (list, optional) features the function is applied to. Overwrites features above
        :param metrics: (list, optional) metrics the function is applied to. Overwrites metrics above
        :param entire: (boolean, optional) apply function to the entire feature's dataframe of metrics?
        :param args: (tuple, optional) args for 'func'
        :param kwargs: (dict, optional) kwargs for 'func'
        """
        # check inputs
        if features is None:
            features = []
        if metrics is None:
            metrics = []
        if not callable(func):
            raise TypeError("functions in ApplyFunc must be callable objects")
        if suffix is not None and not isinstance(suffix, str):
            raise TypeError("prefix, and suffix in ApplyFunc must be strings or None.")
        if prefix is not None and not isinstance(prefix, str):
            raise TypeError("prefix, and suffix in ApplyFunc must be strings or None.")
        if not isinstance(metrics, list) or not isinstance(features, list):
            raise TypeError("metrics and features must be lists of strings.")

        # add function
        self.apply_funcs.append(
            {
                "features": features,
                "metrics": metrics,
                "func": func,
                "entire": entire,
                "suffix": suffix,
                "prefix": prefix,
                "args": args,
                "kwargs": kwargs,
            }
        )

    def transform(
        self,
        apply_to_data: dict,
        assign_to_data: dict | None = None,
        apply_funcs: list | None = None,
    ):
        """
        Apply functions to specified feature and metrics

        Each feature/metric combination is treated as a pandas series

        :param datastore: input datastore
        :return: updated datastore
        :rtype: dict
        """
        assert isinstance(apply_to_data, dict)
        if assign_to_data is None:
            assign_to_data = {}

        if apply_funcs is not None:
            self.apply_funcs += apply_funcs

        if self.msg:
            self.logger.info(self.msg)

        features = self.get_features(list(apply_to_data.keys()))

        same_key = self.assign_to_key == self.apply_to_key

        args = [
            {
                "feature": feature,
                "metrics": self.metrics,
                "apply_to_df": self.get_datastore_object(
                    apply_to_data, feature, dtype=pd.DataFrame
                ),
                "assign_to_df": (
                    None
                    if same_key
                    else self.get_datastore_object(
                        assign_to_data,
                        feature,
                        dtype=pd.DataFrame,
                        default=pd.DataFrame(),
                    )
                ),
                "apply_funcs": self.apply_funcs,
                "same_key": same_key,
            }
            for feature in features
        ]
        result = parallel(apply_func_array, args, mode="kwargs")
        new_metrics = dict(result)
        return new_metrics


def apply_func_array(
    feature, metrics, apply_to_df, assign_to_df, apply_funcs, same_key
):
    """Apply list of functions to dataframe

    Split off for parallelization reasons

    :param str feature: feature currently looping over
    :param list metrics: list of selected metrics to apply functions to
    :param apply_to_df: pandas data frame that function in arr is applied to
    :param assign_to_df: pandas data frame the output of function is assigned to
    :param apply_funcs: list of functions to apply to
    :param same_key: if True, merge apply_to_df and assign_to_df before returning assign_to_df
    :return: union of feature and assign_to_df
    """
    if not isinstance(apply_to_df, pd.DataFrame):
        raise TypeError(
            f'apply_to_df of feature "{feature}" is not a pandas dataframe.'
        )

    if same_key or (len(assign_to_df.index) == 0 and len(assign_to_df.columns) == 0):
        assign_to_df = pd.DataFrame(index=apply_to_df.index)

    for arr in apply_funcs:
        obj = apply_func(feature, metrics, apply_to_df, arr)
        if len(obj) == 0:
            # no metrics were found in apply_to_df
            continue
        for new_metric, o in obj.items():
            if isinstance(o, pd.Series):
                if len(assign_to_df.index) == len(o) and all(
                    assign_to_df.index == o.index
                ):
                    assign_to_df[new_metric] = o
                else:
                    warnings.warn(
                        f"{feature}:{new_metric}: df_out and object have inconsistent lengths."
                    )
            else:
                # o is number or object, assign to every element of new column
                assign_to_df[new_metric] = [o] * len(assign_to_df.index)
    if same_key:
        assign_to_df = pd.concat([apply_to_df, assign_to_df], axis=1)
    return feature, assign_to_df


def apply_func(feature, selected_metrics, df, arr):
    """Apply function to dataframe

    :param str feature: feature currently looping over
    :param list selected_metrics: list of selected metrics to apply to
    :param df: pandas data frame that function in arr is applied to
    :param dict arr: dictionary containing the function to be applied to pandas dataframe.
    :return: dictionary with outputs of applied-to metric pd.Series
    """
    # basic checks of feature
    if (
        "features" in arr
        and len(arr["features"]) > 0
        and feature not in arr["features"]
    ):
        return {}

    # get func input
    keys = list(arr.keys())

    assert "func" in keys, "function input is insufficient."
    func = arr["func"]

    if "prefix" not in keys or arr["prefix"] is None:
        arr["prefix"] = ""
    if len(arr["prefix"]) > 0 and not arr["prefix"].endswith("_"):
        arr["prefix"] = arr["prefix"] + "_"
    if "suffix" not in keys or arr["suffix"] is None:
        arr["suffix"] = func.__name__ if len(arr["prefix"]) == 0 else ""
    if len(arr["suffix"]) > 0 and not arr["suffix"].startswith("_"):
        arr["suffix"] = "_" + arr["suffix"]
    suffix = arr["suffix"]
    prefix = arr["prefix"]

    args = ()
    kwargs = {}
    if "kwargs" in keys:
        kwargs = arr["kwargs"]
    if "args" in keys:
        args = arr["args"]

    # apply func
    if len(selected_metrics) > 0 or ("metrics" in keys and len(arr["metrics"]) > 0):
        metrics = (
            arr["metrics"]
            if ("metrics" in keys and len(arr["metrics"]) > 0)
            else selected_metrics
        )
        metrics = [m for m in metrics if m in df.columns]
        if len(metrics) == 0:
            return {}
        df = df[metrics] if len(metrics) >= 2 else df[metrics[0]]

    if (
        "entire" in arr
        and arr["entire"] is not None
        and arr["entire"] is not False
        and arr["entire"] != 0
    ):
        obj = func(df, *args, **kwargs)
    else:
        obj = df.apply(func, args=args, **kwargs)

    # convert object to dict format
    if not isinstance(
        obj, (pd.Series, pd.DataFrame, list, tuple, np.ndarray)
    ) and isinstance(df, pd.Series):
        obj = {df.name: obj}
    elif not isinstance(
        obj, (pd.Series, pd.DataFrame, list, tuple, np.ndarray)
    ) and isinstance(df, pd.DataFrame):
        obj = {"_".join(df.columns): obj}
    elif (
        isinstance(obj, (list, tuple, np.ndarray))
        and isinstance(df, pd.DataFrame)
        and len(df.columns) == len(obj)
    ):
        obj = dict(zip(df.columns, obj))
    elif (
        isinstance(obj, (list, tuple, np.ndarray))
        and isinstance(df, pd.Series)
        and len(df.index) == len(obj)
    ):
        obj = {df.name: pd.Series(data=obj, index=df.index)}
    elif (
        isinstance(obj, (list, tuple, np.ndarray))
        and isinstance(df, pd.DataFrame)
        and len(df.index) == len(obj)
    ):
        obj = {"_".join(df.columns): pd.Series(data=obj, index=df.index)}
    elif (
        isinstance(obj, pd.Series)
        and isinstance(df, pd.Series)
        and len(obj) == len(df)
        and all(obj.index == df.index)
    ):
        obj = {df.name: obj}
    elif (
        isinstance(obj, pd.Series)
        and isinstance(df, pd.DataFrame)
        and len(obj) == len(df)
        and all(obj.index == df.index)
    ):
        obj = {"_".join(df.columns): obj}
    # e.g. output of normalized_hist_mean_cov: a dataframe with one column, actually a series
    elif (
        isinstance(obj, pd.DataFrame)
        and len(obj.columns) == 1
        and (
            len(obj.index) != len(df.index)
            or (len(obj.index) == len(df.index) and (obj.index != df.index).any())
        )
    ):
        obj = obj[obj.columns[0]].to_dict()
    elif isinstance(obj, pd.Series):
        # e.g. output of np.mean of np.std: results in one number per column when applied to a dataframe
        obj = obj.to_dict()
    elif isinstance(obj, pd.DataFrame):
        # when apply to one column returns another column
        obj = {c: obj[c] for c in obj.columns}
    assert isinstance(obj, dict)

    # add prefix and suffix to dict keys
    keys = list(obj.keys())
    for k in keys:
        obj[prefix + k + suffix] = obj.pop(k)

    return obj



# Copyright (c) 2023 ING Analytics Wholesale Banking
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import numpy as np
import pandas as pd
from tqdm import tqdm

from popmon.base import Module
from popmon.config import Report
from popmon.utils import filter_metrics, short_date
from popmon.visualization.utils import (
    _prune,
    plot_traffic_lights_alerts_aggregate,
    plot_traffic_lights_overview,
)


class TrafficLightSectionGenerator(Module):
    """This module takes the time-series data of already computed statistics, plots the data and
    combines all the plots into a list which is stored together with the section name in a dictionary
    which later will be used for the report generation.
    """

    _input_keys = ("read_key", "dynamic_bounds", "store_key")
    _output_keys = ("store_key",)

    def __init__(
        self,
        read_key,
        store_key,
        settings: Report,
        features=None,
        ignore_features=None,
        static_bounds=None,
        dynamic_bounds=None,
        prefix: str = "traffic_light_",
        suffices=None,
        ignore_stat_endswith=None,
    ) -> None:
        """Initialize an instance of SectionGenerator.

        :param str read_key: key of input data to read from the datastore and use for plotting
        :param str store_key: key for output data to be stored in the datastore
        :param list features: list of features to pick up from input data (optional)
        :param list ignore_features: ignore list of features, if present (optional)
        :param str static_bounds: key to static traffic light bounds key in datastore (optional)
        :param str dynamic_bounds: key to dynamic traffic light bounds key in datastore (optional)
        :param str prefix: dynamic traffic light prefix. default is ``'traffic_light_'`` (optional)
        :param str suffices: dynamic traffic light suffices. (optional)
        :param list ignore_stat_endswith: ignore stats ending with any of list of suffices. (optional)
        """
        super().__init__()
        self.read_key = read_key
        self.store_key = store_key
        self.dynamic_bounds = dynamic_bounds
        self.static_bounds = static_bounds

        self.features = features or []
        self.ignore_features = ignore_features or []
        self.last_n = settings.last_n
        self.skip_first_n = settings.skip_first_n
        self.skip_last_n = settings.skip_last_n
        self.prefix = prefix
        self.suffices = suffices or [
            "_red_high",
            "_yellow_high",
            "_yellow_low",
            "_red_low",
        ]
        self.ignore_stat_endswith = ignore_stat_endswith or []
        self.show_stats = settings.show_stats if not settings.extended_report else None

        self.section_name = settings.section.traffic_lights.name
        self.description = settings.section.traffic_lights.description
        self.tl_colors = settings.tl_colors

    def get_description(self):
        return self.section_name

    def transform(
        self,
        data_obj: dict,
        dynamic_bounds: dict | None = None,
        sections: list | None = None,
    ):
        assert isinstance(data_obj, dict)
        if dynamic_bounds is None:
            dynamic_bounds = {}
        assert isinstance(dynamic_bounds, dict)
        if sections is None:
            sections = []
        assert isinstance(sections, list)

        features = self.get_features(list(data_obj.keys()))
        features_w_metrics = []

        self.logger.info(f'Generating section "{self.section_name}"')

        for feature in tqdm(features, ncols=100):
            df = data_obj.get(feature, pd.DataFrame())
            fdbounds = dynamic_bounds.get(feature, pd.DataFrame(index=df.index))

            assert all(df.index == fdbounds.index)

            # prepare date labels
            df = df.drop(
                columns=["histogram", "reference_histogram"],
                errors="ignore",
            )
            dates = [short_date(date) for date in df.index.tolist()]

            metrics = sorted(
                filter_metrics(df.columns, self.ignore_stat_endswith, self.show_stats)
            )

            plots = [
                _plot_metrics(
                    feature,
                    metrics,
                    dates,
                    df,
                    self.last_n,
                    self.skip_first_n,
                    self.skip_last_n,
                    tl_colors=self.tl_colors,
                )
            ]

            # filter out potential empty plots (from skip empty plots)
            plots = [e for e in plots if len(e["plot"])]

            features_w_metrics.append(
                {
                    "name": feature,
                    "plot_type_layouts": {"traffic_lights": ""},
                    "plots": sorted(plots, key=lambda plot: plot["name"]),
                }
            )

        sections.append(
            {
                "section_title": self.section_name,
                "section_description": self.description,
                "features": features_w_metrics,
            }
        )
        return sections


def _plot_metrics(
    feature,
    metrics,
    dates,
    df,
    last_n,
    skip_first_n,
    skip_last_n,
    tl_colors,
    style: str = "heatmap",
):
    # prune dates and values
    dates = _prune(dates, last_n, skip_first_n, skip_last_n)

    values = [
        _prune(df[metric], last_n, skip_first_n, skip_last_n) for metric in metrics
    ]

    if len(values) > 0:
        values = np.stack(values)

        if style == "heatmap":
            plot = plot_traffic_lights_overview(
                feature, values, metrics=metrics, labels=dates
            )
        elif style == "alerts":
            plot = plot_traffic_lights_alerts_aggregate(
                feature,
                values,
                metrics=metrics,
                labels=dates,
                tl_colors=tl_colors,
            )
        else:
            raise ValueError("style must be either 'heatmap' or 'alerts'")
    else:
        plot = ""

    return {
        "name": "Overview",
        "type": "traffic_light",
        "description": "",
        "plot": plot,
        "full_width": True,
    }



# Copyright (c) 2023 ING Analytics Wholesale Banking
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import annotations

import fnmatch
from textwrap import shorten
from typing import Callable

import pandas as pd
from joblib import Parallel, delayed


def short_date(date):
    """
    Shorten date string to length of 22
    """
    if isinstance(date, pd.Timestamp):
        # Drop the time of day when midnight or noon
        d = (
            str(date).split(" ")[0]
            if date.hour in [0, 12] and date.minute == 0 and date.second == 0
            else str(date)
        )
    else:
        d = str(date)

    return shorten(d, width=22, placeholder="")


def filter_metrics(
    metrics, ignore_stat_endswith: list[str], show_stats: list[str] | None = None
):
    """
    Filter metrics by excluding based on suffix and/or including when matching a pattern
    """
    metrics = [
        m for m in metrics if not any(m.endswith(s) for s in ignore_stat_endswith)
    ]
    if show_stats is not None:
        metrics = [
            m
            for m in metrics
            if any(fnmatch.fnmatch(m, pattern) for pattern in show_stats)
        ]
    return metrics


def parallel(func: Callable, args_list, mode: str = "args"):
    """
    Routine for parallel processing
    """
    from popmon.config import parallel_args

    if parallel_args["n_jobs"] == 1:
        results = [
            func(*args) if mode == "args" else func(**args) for args in args_list
        ]
    else:
        results = Parallel(**parallel_args)(
            delayed(func)(*args) if mode == "args" else delayed(func)(**args)
            for args in args_list
        )
    return results



# Copyright (c) 2023 ING Analytics Wholesale Banking
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
from __future__ import annotations

import fnmatch

import numpy as np
import pandas as pd

from popmon.base import Module


class AlertsSummary(Module):
    """The module AlertsSummary combines the alerts-summaries of all individual features

    It combines the alerts-summaries of all individual features into an artificial feature "_AGGREGATE_".
    """

    _input_keys = ("read_key",)
    _output_keys = ("store_key",)

    def __init__(
        self,
        read_key,
        store_key: str = "",
        features=None,
        ignore_features=None,
        combined_variable: str = "_AGGREGATE_",
    ) -> None:
        """Initialize an instance of AlertsSummary module.

        :param str read_key: key of input data to read from datastore.
        :param str store_key: key of output data to store in datastore (optional).
        :param str combined_variable: name of artificial variable that combines all alerts. default is '_AGGREGATE_'.
        :param list features: features of data frames to pick up from input data (optional).
        :param list ignore_features: list of features to ignore (optional).
        """
        super().__init__()
        self.read_key = read_key
        self.store_key = store_key or self.read_key
        self.features = features or []
        self.ignore_features = ignore_features or []
        self.combined_variable = combined_variable

    def transform(self, data: dict) -> dict | None:
        # determine all possible features, used for the comparison below
        features = self.get_features(list(data.keys()))
        if len(features) == 0:
            return None

        self.logger.info(
            f'Combining alerts into artificial variable "{self.combined_variable}"'
        )

        # STEP 1: loop over features where alerts exist
        df_list = []
        for feature in features:
            # basic checks if feature object is filled correctly
            df = (self.get_datastore_object(data, feature, dtype=pd.DataFrame)).copy(
                deep=False
            )
            df.columns = [f"{feature}_{c}" for c in df.columns]
            df_list.append(df)

        # the different features could technically have different indices.
        # will only merge alerts if all indices are the same
        if len(df_list) >= 2:
            for df in df_list[1:]:
                if not np.array_equal(df_list[0].index, df.index):
                    self.logger.warning(
                        "indices of features are different. no alerts summary generated."
                    )
                    return None

        # STEP 2: Concatenate the dataframes, there was one for each original feature.
        tlv = pd.concat(df_list, axis=1)
        dfc = pd.DataFrame(index=tlv.index)

        # colors of traffic lights
        for color in ["green", "yellow", "red"]:
            cols = fnmatch.filter(tlv.columns, f"*_n_{color}")
            dfc[f"n_{color}"] = tlv[cols].values.sum(axis=1) if len(cols) else 0

        # store combination of traffic alerts
        data[self.combined_variable] = dfc
        return data
